/*
 * FILENAME: osportco.c
 *
 * Copyright  2004 By InterNiche Technologies Inc. All rights reserved
 *
 * Code to map NicheTask "TK_" macros and other port routines
 * to uCOS-II RTOS on ALTERA Cyclone board with the ALTERA Nios2 Core.
 *
 * This file for:
 *   ALTERA Cyclone Dev board with the ALTERA Nios2 Core.
 *   SMSC91C11 10/100 Ethernet Controller
 *   GNU C Compiler provided by ALTERA Quartus Toolset.
 *   Quartus HAL BSP
 *   uCOS-II RTOS Rel 2.76 as distributed by Altera/NIOS2
 *
 * MODULE  : NIOS2GCC
 * PORTABLE: no
 *
 * 06/21/2004
 *
 */

#include "includes.h"      /* from Nios2 uCOS-II directory */

#include "ipport.h"        /* from Interniche directory */

#ifdef UCOS_II             /* Whole file is #ifdef'ed away if no uCOS-II */

#include "osport.h"

#include "in_utils.h"
#include "memwrap.h"

void *   net_task_sem_ptr = NULL;

extern OS_EVENT * mheap_sem_ptr;

/* state of ints, saved by XXlock_net_res() */
int   netq_intmask;

INT8U    netmain_prio = 0;    /* ID of netmain task, for SignalPktDemux() */

extern int netmain(void);

#ifdef NOT_USED
char mempool[0x60000];
#endif

int old_mode;

unsigned long cticks;

u_char
uCOS_self(void)
{
   /* Return uCOS currently running task ID */
   return(OSTCBCur->OSTCBPrio);
}


extern void irq_Mask(void);
extern void irq_Unmask(void);

void
LOCK_NET_RESOURCE(int resid)
{
INT8U error;
extern void irq_Mask(void);

   error = 0;

   switch(resid)
   {
   case NET_RESID:
      OSSemPend(net_task_sem_ptr, 0, &error);

      if(error)
      {
         int errct = 0;

         /* sometimes we get a "timeout" error even though we passed a zero
          * to indicate we'll wait forever. When this happens, try again:
          */
         while(error == 10)
         {
            if(errct++ > 1000)
            {
               panic("lock NET");    /* fatal */
               return;
            }
            OSSemPend((OS_EVENT *)net_task_sem_ptr, 0, &error);
         }
      }
      break;
   case RXQ_RESID:
      irq_Mask();
      break;
   case FREEQ_RESID:
      irq_Mask();
      break;
   default:
      dtrap();   /* unknown resource id passed to LOCK_NET_RESOURCE */
      break;
   }
}

void
UNLOCK_NET_RESOURCE(int resid)
{
INT8U error;
extern void irq_Unmask(void);

   error = 0;

   switch((int)resid) 
   {
   case NET_RESID:
      error = OSSemPost((OS_EVENT *)net_task_sem_ptr);
      if(error != OS_NO_ERR)
         panic("unlock NET");
      break;
   case RXQ_RESID:
      irq_Unmask();
      break;
   case FREEQ_RESID:
      irq_Unmask();
      break;
   default:
      dtrap();   /* unknown resource id passed to UNLOCK_NET_RESOURCE */
      break;
   }
}


/*
 * Here's the workaround to an unusual quirk of uCOS-II - It
 * cannot have multiple tasks of the same priority. Networking systems
 * are best done with a few threads which have equal priority and thus 
 * ensure fair CPU sharing without convoluted priority balancing. This
 * means we have to guess a priority for each of our net tasks and then
 * code tk_yield() to force a non-zero delay (one tick), to ensure that all
 * the network tasks get a round-robin slot to do their thing. Of course
 * we could alter the code's design to be optimized for a strictly
 * prioritized RTOS, but that would penalize all the non-preemptive and
 * non-prioritized systems we also support.
 */

extern TK_ENTRY(tk_netmain);        /* in netmain.c */
extern long     netmain_wakes;

#ifdef TK_STDIN_DEVICE
extern TK_ENTRY(tk_keyboard);       /* in netmain.c */
extern long     keyboard_wakes;
#endif

extern TK_ENTRY(tk_nettick);        /* in netmain.c */
extern long     nettick_wakes;

int
TK_NEWTASK(struct inet_taskinfo * nettask)
{
   INT8U    error;
   OS_STK * stack;

   stack = (OS_STK*)npalloc(nettask->stacksize);
   if(!stack)
      panic("stack alloc");

#if OS_TASK_CREATE_EXT_EN > 0
   error = OSTaskCreateExt(
      nettask->entry,
      nettask->task_parameter,
      stack + (nettask->stacksize/sizeof(OS_STK)) - 1,
      nettask->priority,
      nettask->priority,
      stack, 
      (INT32U)nettask->stacksize / sizeof(OS_STK),
      NULL,
      OS_TASK_OPT_STK_CHK | OS_TASK_OPT_STK_CLR);
#else
   error = OSTaskCreate(
      (void (*)(void *)) nettask->entry,
      nettask->task_parameter,
      stack + (nettask->stacksize/sizeof(OS_STK)) - 1,
      nettask->priority);
#endif
   /* If we go here, then there's another task using our priority */
   /* Tell the user and exit with an error */
   if(error == OS_PRIO_EXIST)
   { 
     char curr_task[OS_TASK_NAME_SIZE];
     INT8U err;
     OSTaskNameGet(nettask->priority, curr_task, &err);
     curr_task[OS_TASK_NAME_SIZE-1]=0;
    
     printf("Priority requested for task \"%s\" (Prio:%d) conflicts with "\
            "already running task \"%s\" (Prio: %d)\n",
             nettask->name, nettask->priority, curr_task, nettask->priority);

     printf("You may wish to check your task priority settings in "\
            "\"Debug\\system_description\\ipport.h\" against "\
            "the priority settings in your application and recompile.\n\n");
   }
   else if(error == OS_PRIO_INVALID)
   {
     printf("Priority requested for task \"%s\" (Prio:%d) exceeds "\
            "available priority levels in the system (OS_LOWEST_PRIO = %d)\n\n",
             nettask->name, nettask->priority, OS_LOWEST_PRIO);
             
     printf("Please modify the tasks priority level, or modify the "\
            "\"Lowest assignable priority\" setting in the MicroC/OS-II "\
            "component\n");
   }

   /* All other errors are fatal */
   if(error)
   {
      printf("Task create error /(MicroC/OS-II error code:%d/) on %s\n", error, nettask->name);
      return -1;
   }

   /* Include the task name, so that uc/osII (os aware) debuggers can
    * display it.
    */
   OSTaskNameSet(nettask->priority, &nettask->name[0], &error);

   nettask->stackbase = (char*)stack;
   nettask->priority = nettask->priority;
   *nettask->tk_ptr = (INT8U)nettask->priority;  

   printf("Created \"%s\" task (Prio: %d)\n",
         (char *)nettask->name, nettask->priority);

   /* Remember netmain task Id/Priority for SignalPktDemux() */
   if(nettask->entry == tk_netmain)
      netmain_prio = nettask->priority;

   return 0;
}


extern void OSTimeTick (void);


/* Rest of file is for TEST Purposes Only and is excluded unless we're
 * doing lock_net checking.
 */

#ifdef LOCKNET_CHECKING
#include "q.h"       /* InterNiche queue defines */

/* locally define external items involved in checking locks */
extern queue rcvdq;
extern queue bigfreeq;
extern queue lilfreeq;
extern int iniche_net_ready;
extern queue mfreeq;

extern int irq_level;  /* level of nesting in irq_Mask() and irq_Unmask() */

/* FUNCTION: LOCKNET_CHECK()
 *
 * If the passed queue is one of the sensitive ones then make sure
 * the related resource is locked.
 *
 * PARAM1: queue * q
 *
 * RETURNS: nothing, panics if test fails
 */

void
LOCKNET_CHECK(struct queue * q)
{

   /* Don't start checking until the net is up */
   if(!iniche_net_ready)
      return;

#ifdef INCLUDE_TCP
#ifndef MINI_TCP
   /* On uCOS/Nichestack/TCP builds, check the mbuf free queue is locked
    * by the NET_RESID semaphore. If it's not we panic().
    */
   if(q == &mfreeq)
   {
      /* A non-zero from OSSemAccept() means the semaphore was NOT locked */
      if(OSSemAccept(net_task_sem_ptr) != 0)
      {
         panic("locknet_check1");
      }
   }
#endif /* not MINI_TCP */
#endif /* INCLUDE_TCP */

   /* quick return if it's not one of the other sensitive queues */
   if((q != &rcvdq) &&
      (q != &bigfreeq) &&
      (q != &lilfreeq))
   {
      return;
   }

   /* Since the Nios2 build uses interrupt disabling to protect these
    * queues, we just need to check the Interrupt state. We use look at
    * the irq_level to get the current interrupt state, if interrupts
    * were not then we are going to panic().
    */

   if(irq_level != 1)    /* Get current interupt state */
   {
      panic("locknet_check2");
   }
   return;
}

#endif /* LOCKNET_CHECKING */

#endif /* UCOS_II */

