// Message.h: interface for the CNQMessage class.
//
//////////////////////////////////////////////////////////////////////

#ifndef __NQMESSAGE_H_INCLUDED_SW__
#define __NQMESSAGE_H_INCLUDED_SW__


/** \brief platform independent message object, used as transport buffer for the NetworkQueue
    \date SW (Updated August 2007)           
	\section notes Important Notes
		- there exist public variables one can write to 
		- inside the CNQMessage packet buffer all payload_buffer contents are _always_ in little endian format
		  for implementation reasons. 
		- On the network, any data that is put into the UDP packet is encoded in little endian format
		- The payload_buffer is a linked list type, the structure is as follows:
			- <b>UINT16 next</b>         offset to next entry
			- <b>UINT16 gross_size</b>	 sizeof data block (id+type+grosssiz+numele+data)			
			- <b>UINT16 id</b>           data id (to be defined by application)
			- <b>UINT16 type</b>         data type (see NQDeclarations.h VARTYPE stuff)
			- <b>UINT32 num_elemen</b>   num of elements of data type
			- ...... net data of this block until next
            .
        - Next Packet:
            - <b>UINT16 next</b>         offset to next entry
            - <b>UINT16 gross_size</b>   sizeof data block (id+type+grosssiz+numele+data)           
            - <b>UINT16 id</b>           data id (to be defined by application)
            - <b>UINT16 type</b>         data type (see NQDeclarations.h VARTYPE stuff)
            - <b>UINT32 num_elemen</b>   num of elements of data type
            - ..... net data of this block until n_next
            .
        - Next Packet ... 
    \section platform Platform dependend data encoding
	For performance and MMU / absence of MMU reasons on RISC platform, any 'net data' in the block 
	with less than four bytes is padded to four bytes.

	This is because of Sparc core dump if unaligned access to e.g. 4-byte integer.
	This is because NIOS2 processor (little endian RISC) does not contain an MMU and exhibits the
	same behaviour as Sparc CPU, but there is no 'core dump' but unaligned memory access is shifted
	to the last/next address (can't remember correctly) - so basically one is writing at a wrong 
	location in memory - possibly overwriting data!

	Fragmentation may happen the more times data is overwritten by bigger data!
	There is no reorganization in this linked list if gaps are formed.            
*/
/*
        - CNQMessage and CNetworkQueue are friends !
        - CNQMessage and CNQPacketQueue are friends !   
 */
class CNQMessage
{
	friend class CNetworkQueue; // (---)
	friend class CNQPacketQueue; // (-x-)
public:
	// methods

	//! Constructor
	CNQMessage();
    //! Destructor
	~CNQMessage();

	//! reset contents of message
	void clear();

    //! \brief check for valid public fields
    //! \return exact error code
	INT32 check_int() const;
    
    //! \brief check for valid public fields 
    //! \retval true public fields are valid
	bool check_bool() const; // valid public fields (?)

    //! \brief check, if message is a response packet
    //! \retval true packet is a response paket
	bool isResponse() const;

    //! \brief check, if message is a request packet
    //! \retval true packet is a request paket
    bool isRequest() const;

    //! \brief check, if message is a telegram packet
    //! \retval true acket is a telegram paket
	bool isTelegram() const;

    //! \brief get Sender IP and Port from CNQMessage 
    //! \retval true successfully returned requested data
    //! \retval false the message has no valid sender information (it did not come from remote peer) 
	bool getSenderIPandPort( UINT32 &ip_peer_hostorder, UINT16 &port_peer_hostorder );

    // read-API   
    //@{
    //! \name reading data
    
    /*! \brief read a binary object from the CNQMessage Object
        \param varid numeric identifier of a packet field inside the message
        \param target_buffer target buffer to copy data to
        \param target_buffer_size_bytes target buffer size, if to small nessesary value is written back and function returns false
        \return true on success
    */
	bool read_BINARY( UINT16 varid, UINT8 *target_buffer, UINT32 &target_buffer_size_bytes);
    /*! \brief read a float value from the CNQMessage Object
        \param varid numeric identifier of a packet field inside the message
        \param target_buffer target buffer to copy data to
        \return true on success
    */
    bool read_FLOAT( UINT16 varid, FLOAT &target_buffer);
    /*! \brief read a double float value from the CNQMessage Object
        \param varid numeric identifier of a packet field inside the message
        \param target_buffer target buffer to copy data to
        \return true on success
    */
	bool read_DOUBLE( UINT16 varid, DOUBLE &target_buffer);
    /*! \brief read a 8 bit signed value from the CNQMessage Object
        \param varid numeric identifier of a packet field inside the message
        \param target_buffer target buffer to copy data to
        \return true on success
    */    
	bool read_INT8( UINT16 varid, INT8 &target_buffer);
    /*! \brief read a 8 bit unsigned value from the CNQMessage Object
        \param varid numeric identifier of a packet field inside the message
        \param target_buffer target buffer to copy data to
        \return true on success
    */    
	bool read_UINT8( UINT16 varid, UINT8 &target_buffer);
    /*! \brief read a 16 bit signed value from the CNQMessage Object
        \param varid numeric identifier of a packet field inside the message
        \param target_buffer target buffer to copy data to
        \return true on success
    */    
	bool read_INT16( UINT16 varid, INT16 &target_buffer);
    /*! \brief read a 16 bit unsigned value from the CNQMessage Object
        \param varid numeric identifier of a packet field inside the message
        \param target_buffer target buffer to copy data to
        \return true on success
    */    
	bool read_UINT16( UINT16 varid, UINT16 &target_buffer);
    /*! \brief read a 32 bit signed value from the CNQMessage Object
        \param varid numeric identifier of a packet field inside the message
        \param target_buffer target buffer to copy data to
        \return true on success
    */    
	bool read_INT32( UINT16 varid, INT32 &target_buffer);
    /*! \brief read a 32 bit unsigned value from the CNQMessage Object
        \param varid numeric identifier of a packet field inside the message
        \param target_buffer target buffer to copy data to
        \return true on success
    */    
	bool read_UINT32( UINT16 varid, UINT32 &target_buffer);
    /*! \brief read a text string from the CNQMessage Object
        \param varid numeric identifier of a packet field inside the message
        \param target_buffer target buffer to copy data to
        \param target_buffer_size_bytes target buffer size, if to small nessesary value is written back and function returns false
        \return true on success
    */    
	bool read_TEXT( UINT16 varid, UINT8 *target_buffer, UINT32 &target_buffer_size_bytes);
    /*! \brief general read function for non standard var types 
        \param varid            numeric identifier of a packet field inside the message
        \param vartype          numeric identifier of the var type 
        \param target_buffer                target buffer to copy data to
        \param target_buffer_num_elements   number of target buffer elements, if to 
                                            small nessesary value is written back and 
                                            function returns false
        \return true on success
    */    
	bool read( UINT16 varid, UINT16 vartype, VOIDP target_buffer, UINT32 &target_buffer_num_elements);

    //@}

	// write-API
    //@{
    //! \name writinge data

    /*! \brief write a binary object into the CNQMessage Object
        \param varid numeric identifier of a packet field inside the message
        \param source_buffer source buffer to copy data from
        \param source_buffer_size_bytes size of the source buffer in bytes
        \return true on success
        \attention if varid already exists, existing content is overwritten 
     */
    bool write_BINARY( UINT16 varid, const UINT8 *source_buffer, UINT32 source_buffer_size_bytes );
    /*! \brief write a float value into the CNQMessage Object
        \param varid numeric identifier of a packet field inside the message
        \param source_buffer source buffer to copy data from
        \return true on success
        \attention if varid already exists, existing content is overwritten 
     */
	bool write_FLOAT( UINT16 varid, FLOAT source_buffer);
    /*! \brief write a double float value into the CNQMessage Object
        \param varid numeric identifier of a packet field inside the message
        \param source_buffer source buffer to copy data from
        \return true on success
        \attention if varid already exists, existing content is overwritten 
     */
	bool write_DOUBLE( UINT16 varid, DOUBLE source_buffer);
    /*! \brief write a signed 8bit value into the CNQMessage Object
        \param varid numeric identifier of a packet field inside the message
        \param source_buffer source buffer to copy data from
        \return true on success
        \attention if varid already exists, existing content is overwritten 
     */
	bool write_INT8( UINT16 varid, INT8 source_buffer);
    /*! \brief write a unsigned 8bit value into the CNQMessage Object
        \param varid numeric identifier of a packet field inside the message
        \param source_buffer source buffer to copy data from
        \return true on success
        \attention if varid already exists, existing content is overwritten 
     */
	bool write_UINT8( UINT16 varid, UINT8 source_buffer);
    /*! \brief write a signed 16bit value into the CNQMessage Object
        \param varid numeric identifier of a packet field inside the message
        \param source_buffer source buffer to copy data from
        \return true on success
        \attention if varid already exists, existing content is overwritten 
     */
	bool write_INT16( UINT16 varid, INT16 source_buffer);
    /*! \brief write a unsigned 16bit value into the CNQMessage Object
        \param varid numeric identifier of a packet field inside the message
        \param source_buffer source buffer to copy data from
        \return true on success
        \attention if varid already exists, existing content is overwritten 
     */
	bool write_UINT16( UINT16 varid, UINT16 source_buffer);
    /*! \brief write a signed 32bit value into the CNQMessage Object
        \param varid numeric identifier of a packet field inside the message
        \param source_buffer source buffer to copy data from
        \return true on success
        \attention if varid already exists, existing content is overwritten 
     */
	bool write_INT32( UINT16 varid, INT32 source_buffer);
    /*! \brief write a unsigned 32bit value into the CNQMessage Object
        \param varid numeric identifier of a packet field inside the message
        \param source_buffer source buffer to copy data from
        \return true on success
        \attention if varid already exists, existing content is overwritten 
     */
	bool write_UINT32( UINT16 varid, UINT32 source_buffer);
    /*! \brief write a text string from the CNQMessage Object
        \param varid numeric identifier of a packet field inside the message
        \param source_buffer source buffer to copy data from
        \param source_buffer_size_bytes size of source buffer in bytes
        \return true on success
        \attention if varid already exists, existing content is overwritten 
     */
	bool write_TEXT( UINT16 varid, const UINT8 *source_buffer, UINT32 source_buffer_size_bytes);
    /*! \brief general write function for non standard var types 
        \param varid        numeric identifier of a packet field inside the message
        \param vartype      numeric constant that determines the var type (see #NQ_VARTYPE_BINARY, #NQ_VARTYPE_FLOAT ...)
        \param buffer       source buffer to copy data from
        \param elements_in_buffer   number of elements in buffer 
        \return true on success
        \attention if varid already exists, existing content is overwritten 
     */
	bool write( UINT16 varid, UINT16 vartype, const VOIDP buffer, const UINT32 elements_in_buffer);
    
    //@}

    /*! \brief set physical and logical paket type
     *  \param aPhysicalType numerical identifier of physical paket type
     *  \param aLogicalType numerical identifier of logical paket type
     */
	void  setType( UINT8 aPhysicalType, UINT8 aLogicalType);
    //! get the physical paket type
    UINT8 getPhysicalType();
    //! get the logical paket type
	UINT8 getLogicalType();

public:
	// variables

	// public variables the user programmer may set
    //@{
    //! \name public variables free to use for the user code
    
    //! authorisation ID, an optional parameter to use for paket management in user code
	UINT32	optAuthorisationID;    
    //! subsystem device id, an optional parameter to use for paket management in user code 
	UINT16	optSubsystemDeviceID;  
    
    /*! \brief packet type additional parameter
     *  
     *   - for TRAP type: recommendation: contains TRAP_ID
     *   - for RESP type: recommendation: contains error code
     */    
	UINT16	optPacketTypeParam;	

    //@}
								
private:
	// methods

	// ack-packet handling
	bool matchAckPacket(const CNQMessage &ack_packet);
	void constructAckPacket(const CNQMessage &template_packet);
	
	bool isAckPacket() const;	// true: packet IS an ack packet
	bool wantAckPacket() const; // true: packet wants an ack back from remote peer

	// return size of one element of the passed vartype
	inline UINT32 sizeof_per_element( const UINT16 vartype );

	// used by the network queue to encode and decode CNQMessage packets to UDP packets and vice versa
	// physical layout as binary <-> logical layout on object CNQMessage
	bool ImportMessageFromPacketBuffer(const UINT8 *buffer, INT32 buffer_length, UINT32 ip_hostorder, UINT16 port_hostorder);
	bool ExportMessageToPacketBuffer(UINT8 *buffer, INT32 &buffer_length);

	// copies data from source to target
	// with respect to possible endian changing
	void copy_data_depending_on_vartype( UINT32 *tgtbufpos, const UINT32 *srcbufpos, UINT32 num_elements, UINT16 vartype);

	// correctness check whether the buffer really contains the VARTYPE data
	// mostly size is checked
	bool check_buffer_against_vartype( UINT16 vartype, const VOIDP buffer, UINT32 buflength);

private:
	// packet-content variables

	// public variables the user programmer must set
	UINT8	mandatoryLogicalPacketType;
	UINT8	mandatoryPhysicalPacketType;
	

	INT32	payload_first_offset;	// first element in payload_buffer starts here
	UINT8	payload_buffer[(NQ_MAX_NUM_ENTRIES*NQ_MIN_SIZE_DATA)+NQ_SIZE_ADD_LINKEDLIST];
									// buffer in memory to temporaily hold payload
	
	UINT16	payload_buffer_left;	// space that is left (number of chars) in payload_buffer
								 
									// this will be initialized to 
									// (NQ_MAX_NUM_ENTRIES*NQ_MIN_SIZE_DATA)+NQ_SIZE_ADD_LINKEDLIST

	UINT16	payload_curlength_net;	// current data length as it would occur ON THE NETWORK!
	UINT16	payload_num_elements;	// number of elements in linked list 


	// to be set/get/used by the network queue
	UINT16	internalPacketIndex;			// (---)
	UINT16	internalRequestID;				// (---) // (-x-)
 

	UINT16  payload_maxlength_net;	// initialized to the maximum length a packet can have on the network

	UINT16	internalPacketMagic;
	UINT16	internalPacketSize;
	UINT16	internalPacketVersion;
	UINT16  internalPacketSpecialBits;	// special info bits
				
									// see: Declarations.h PACKET_SPECIAL_BITS_VALID_MASK
								    
									// combination 1-1 is illegal, ACK packet can not want ACK back
									// first bit:  0 - normal packet 1 - ACK packet
									// second bit: 0 - want no ACK   1 - want ACK back
									// 3to16th bit: unused, must be set to zero!



	UINT16	internalPacketTail;

private:
	// internal handling data
	/////////////////////////////////////////////////////////

	bool	message_came_from_set;
	UINT32  message_came_from_ip;
	UINT16  message_came_from_port;
	
	// in host byte order
	// send thread needs that in order to know where to send the message
	UINT32  message_sendto_ip_host;
	UINT16  message_sendto_port_host;


	HANDLE  notifySemaphoreOnReceiveAnswer;	// for blocking synchronous calls									
	HANDLE	notifySemaphoreOnReceiveAck;	// for blocking sends that require ACK packet to return
};

#endif // __NQMESSAGE_H_INCLUDED_SW__
