#ifndef __NETWORKQUEUE_H_INCLUDED__
#define __NETWORKQUEUE_H_INCLUDED__


//! Encapsulates a Networkqueue in a object.
/*! creates a Networkqueue endpoint for communication with a nother Networkqueue endpoint.
 */
class CNetworkQueue
{
	friend class CNQMessage; // (---)
	friend class CNQMessageBuffers; // (-!-)
public:
    //! constructor of the NetworkQueue object
	CNetworkQueue();

    //! destructor of the NetworkQueue object
	~CNetworkQueue();

	// reference handling to get/putback CNQMessage buffers to work with
    /*! \brief create a valid CNQMessage object reference
     * 
     *  The CNQMessage object reference must be freed by CNetworkQueue::freeMsg.
        You must use only this function, to create valid CNQMessage objects, because CNQMessage object references are
        handled internaly in a dedicated memory area, accessable by the asynchronous thread operations. 
            
        \returns valid Reference to a CNQMessage object or \c NULL
        \sa freeMsg
    */
	CNQMessage	*createMsg();

    //! \brief free a CNQMessage reference which has previously been created or returned 
	//! this CNQMessage instance might be returned at CNetworkQueue::createMsg, CNetworkQueue::recvMsg() and
	//! CNetworkQueue::sendRecvMsg()
	//! 
	bool		freeMsg( CNQMessage *ref);
    
    /**  add a new NetworkQueue endpoint to the accept list.
     * 
        \param ip_hostorder     The IP address (in host byte order)
        \param port_hostorder   The Port number (in host byte order)
    */
	bool		addPeer(UINT32 ip_hostorder, UINT32 port_hostorder);
	
    /**  remove a NetworkQueue endpoint from the accept list.
     *  
        \param ip_hostorder     The IP address (in host byte order)
        \param port_hostorder   The Port number (in host byte order)
    */
    bool		removePeer(UINT32 ip_hostorder, UINT32 port_hostorder);
	
    /**  create and connect the Networkqueue to a remote peer
     *  \param ip_remote_peer       The remote peer IP address as a string
     *  \param portnumber_remote    The remote peer port number to send to (in host byte order)
     *  \param portnumber_local     The local port to send from (in host byte order)
     */
    bool		create( const ONECHAR *ip_remote_peer, UINT16 portnumber_remote, UINT16 portnumber_local ); // dotted IP address of remote peer
    // closes the Networkqueue and deinitializes peer connection
	bool		close();

    //!@{
    //! \name       send and receive API       
    
    /**  \brief     Send a message from local Networkqueue immediately to remote peer.
     */
	bool		sendMsgTo( CNQMessage *msg, UINT32 ip_hostorder, UINT16 port_hostorder);

    /** \brief      Send a message from Networkqueue with timeout 
     *  \param      msg             reference to the CNQMessage object to send 
     *  \param      ip_hostorder    the destination peer IP address (in host byte order) 
     *  \param      port_hostorder  the destination peer port address (in host byte order) 
     *  \param      timeout         timeout to wait max. for the next message
     *  \return     true, if a message was successfully send
     */
	bool		sendMsgTo( CNQMessage *msg, UINT32 ip_hostorder, UINT16 port_hostorder, const NQ_TIMEVAL &timeout);

	/**  \brief      Receive/wait for next message from Networkqueue 
     *  \param      newmsg          pointer variable to get the CNQMessage reference of the next received message
     *  \return     true, if a message was received
     *  \attention  You must free the CNQMessage Object with CNetworkQueue::freeMsg after using it!
     */
	bool		recvMsg( CNQMessage **newmsg ); // instant return
    
    /**  \brief      Receive next message from Networkqueue and give wait timeout
     *  \param      newmsg          pointer variable to get the CNQMessage reference of the next received message
     *  \param      timeout         timeout to wait max. for the next message
     *  \return     true, if a message was received
     *  \attention  You \em must free the CNQMessage Object with CNetworkQueue::freeMsg after using it!
     */
	bool		recvMsg( CNQMessage **newmsg, const NQ_TIMEVAL &timeout ); // event-based, wait until timeout

    /**  \brief      Send and receive a message from Networkqueue with timeout (rpc-like call)
     *  \param      response        pointer variable to get the response CNQMessage reference 
     *  \param      request         reference reference to the CNQMessage object to send for the request 
     *  \param      ip_hostorder    the destination peer IP address (in host byte order) 
     *  \param      port_hostorder  the destination peer port address (in host byte order) 
     *  \param      timeout         timeout to wait max. for the next message
     *  \return     true, if a message was received
     *  \attention  You \em must free the CNQMessage Object with CNetworkQueue::freeMsg after using it!
     */
	bool		sendRecvMsg(CNQMessage **response, CNQMessage *request, UINT32 ip_hostorder, UINT16 port_hostorder, const NQ_TIMEVAL &timeout);

    /**  \brief       get statistical information about the NetworkQueue
     *  \param       stats           pointer to a STATISTICS structure to retrieve the stastical information
     *  \return      true, if the statistical information was retrieved successfully
     */ 
	bool		getStatistics( NQ_STATISTICS &stats );

    //!@}
    
    //          get the last error code 
	UINT32		getLastError();
    //          get the error string of the last occured error code
	bool		getLastErrorString( ONECHAR *buffer, UINT32 buflength);
	
	// should not be public in principle!
	// public because called from platform.cpp non-class thread function
	// one task or thread in order to do sending and receiving from peer
    
    /*! \brief  public only by software design reason
     * 
     *          This member function is called internally by the sender thread/task and
	 *			performs NetworkQueue send logic:
     */ 
    void		SendProcessTask();
    
    /*! \brief  public only by software design reason
     * 
     *          This member function is called internally by the receiver thread/task and
	 *			performs NetworkQueue receive logic.
     */ 
	void		ReceiveProcessTask();
	
	//void PrintQueueSizes(); // was debug, removed Oct 17, 2007 SW

protected:

private:

	bool				sendMsgToInternal( UINT32 ip_hostorder, UINT16 port_hostorder, CNQMessage *msg, HANDLE recv_notify_sem);

	CNQMessage			*obtainReference();
	bool				incrementReferenceCount( CNQMessage *ref );
	bool				putbackReference( CNQMessage *ref);

	
	// pooling of message buffers
	CNQMessagePool		bufs;

	// internal statistics buffer
	NQ_STATISTICS		m_stats;

	socktype			m_socket;
	UINT32				m_iLastError;
	INT32				m_iLastSystemError;

	// semaphores to communicate with send and receive tasks/threads	
	HANDLE				m_sendSemaphore;
	HANDLE				m_receiveSemaphore;
	
	// queues to communicate with send and receive tasks/threads
	CNQPacketQueue		m_SendPacketQueue;
	CNQPacketQueue		m_ReceivePacketQueue;

	// dedicated queues for special handling
	CNQPacketQueue		m_AckPacketWaitingQueue;

	CNQPacketQueue		m_ReceiveResponseSemaphorePacketQueue;
	CNQResponseWaitQueue	m_ResponseWaitQueue;

	// list of allowed peers
	CNQPeerList			m_PeerList;

	bool				m_bSuccessfulInitialized;

	// thread/task stuff
	HANDLE				m_hSendThread;
	HANDLE				m_hReceiveThread;

	volatile bool		m_bShutdownSendThread;
	volatile bool		m_bSendThreadRunning;

	volatile bool		m_bShutdownReceiveThread;
	volatile bool		m_bReceiveThreadRunning;

	// 'unique' request id
	UINT16				m_iNextRequestId;
};


#endif //__NETWORKQUEUE_H_INCLUDED__

