// NQMessagePool.cpp: implementation of the CNQMessagePool class.
//
//////////////////////////////////////////////////////////////////////

#include "PlatformSW.h"
#include <assert.h>
#include "NQDeclarations.h"
#include "NQLogging.h"
#include "NQMessage.h"
#include "NQLockMechanism.h"
#include "NQMessagePool.h"

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

//
// constructor (basic variable initialization)
//
CNQMessagePool::CNQMessagePool(): CNQLockMechanism()
{
	m_pCMessages = NULL;
	m_pTakenBuffer = NULL;
	m_bInitialized = false;
	m_iSharedReferences = 0;
	m_iNumBuffers = 0;
}


//
// destructor (mandatory clean-up deinitialize)
//
CNQMessagePool::~CNQMessagePool()
{
	Deinitialize();
}

// CNQMessagePool::Initialize()
/** \brief Initialize 'numbuffers' empty CNQMessage objects
 * 
 * 	This method initializes 'numbuffers' CNQMessage object storage locations to
 *  avoid huge static and stack data.
 * 
 * 	\param numbuffers	amount of CNQMessage buffers in pool
 * 
 *  \retval true successfully initializes to 'numbuffers' buffers in pool
 * 	\retval	false something happened (mutex, already initialized)
 * 
 * 	\note
 * 		This function contains assert() which give more failure reasons during development/debugging.
 */
bool CNQMessagePool::Initialize( UINT32 numbuffers )
{
	if (m_bInitialized == true) return false;

	if (!lock()) return false;

	assert( numbuffers > 0 );

	m_pCMessages = new CNQMessage[numbuffers];
	assert( m_pCMessages != NULL);

	m_pTakenBuffer = new UINT32[numbuffers];
	assert(m_pTakenBuffer != NULL);

	for (UINT32 i=0;i<numbuffers;i++)
	{
		m_pTakenBuffer[i] = 0;
	}
	
	m_bInitialized = true;
	m_iSharedReferences = 0;
	m_iNumBuffers = numbuffers;

	unlock();

	return true;
}

//
// CNQMessagePool::Deinitialize()
/** \brief  Free memory, clean up anything
 * 
 * 	This method deinitializes (frees, cleans up) CNQMessage object storage pool.
 * 
 * 	\retval true successfully deinitialized 
 * 	\retval false something happened, unsuccessful deinitialization (mutex)
 * 				  user programmer: treat it as not deinitialized at all
 * 
 * 	\note
 * 		This function contains assert() which give more failure reasons during development/debugging.
 */
bool CNQMessagePool::Deinitialize()
{
	if (m_bInitialized == true)
	{
		if (!lock()) return false;

		NQ_DEBUGMODE_ONLY(if (m_iSharedReferences>0) NQ_DBG(("clean CNQMessagePool: m_iSharedReferences=%d\n",m_iSharedReferences)));
		assert( m_iSharedReferences == 0 );

		delete [] m_pCMessages;
		delete [] m_pTakenBuffer;

		m_pCMessages = NULL;
		m_bInitialized = false;
		m_iSharedReferences = 0;
		m_iNumBuffers = 0;
		m_pTakenBuffer = NULL;
		
		unlock();
	}

	return true;
}

//
// CNQMessagePool::acquire_reference()
//
/** \brief  Obtain a free (refcount 0) CNQMessage object
 * 
 * 	This method looks around a pool for a 'free' (not already taken) CNQMessage buffer and returns it 
 *  in case it was found. The reference count for that pool element is increased to mark that one (1)
 *  requested it.
 * 
 * 	\retval !=NULL one pool member returned 
 * 	\retval	NULL something happened, no pool member returned obviously (not initialized, mutex or any
 * 				 buffer is taken)
 */
CNQMessage *CNQMessagePool::acquire_reference()
{
	if (m_bInitialized == false) return NULL;

	if (!lock()) return NULL;

	CNQMessage *retval=NULL;
	
	for (UINT32 i=0;i<m_iNumBuffers;i++)
	{
		if (m_pTakenBuffer[i] == 0) 
		{
			m_pTakenBuffer[i] = 1;
			retval = &(m_pCMessages[i]);
			m_iSharedReferences++;
			break;
		}
	}

	unlock();

	return retval; // nothing free or buffer ptr
}

//
// CNQMessagePool::increment_referencecount()
//
/** \brief Increment reference count by 1 for a already taken CNQMessage (by get_reference())
 * 
 * 	This method looks around a pool for the 'buffer_ref' (already taken) CNQMessage buffer and 
 *  in case it was found, the reference count for that pool element is increased by one (1).
 * 
 * 	\param buffer_ref buffer reference already taken
 * 
 * 	\retval	true reference count incremented by one
 * 	\retval	false reference count NOT incremented (no init, mutex, buffer_ref was not found or 
 * 				  not taken at all)
 * 
 * 	\note
 * 		This function contains assert() which give more failure reasons during development/debugging.
 */
bool CNQMessagePool::increment_referencecount( CNQMessage *buffer_ref )
{
	bool retval = false;
	if (m_bInitialized == false) return retval;
	
	if (!lock()) return false;

	for (UINT32 i=0;i<m_iNumBuffers;i++)
	{
		if (&(m_pCMessages[i]) == buffer_ref)
		{
			assert( m_pTakenBuffer[i] > 0);
			if (m_pTakenBuffer > 0) 
			{
				m_pTakenBuffer[i]++;
				retval = true;
			}
			else retval = false;
			
			break;
		}
	}

	unlock();

	return retval; // nothing free or buffer ptr
}


//
// CNQMessagePool::putback_reference()
//
/** \brief Put reference back (decrement reference count and clear CNQMessage if refcount is zero)
 * 
 * 	This method looks around a pool for the 'ref' (already taken) CNQMessage buffer and 
 *  in case it was found, the reference count for that pool element is decreased by one (1).
 * 	After that, it is checked whether the reference count has reached zero (no references)
 *  again. If that is the case, the CNQMessage buffer is cleared. For subsequent calls to 
 *  acquire_reference, this buffer might be delivered again, because no code location is using
 * 	it any more.
 * 
 * 	\param ref buffer reference already taken
 * 
 *  \retval true successfully put back (does NOT mean that this was the last reference so the 
 * 				buffer is available again!)
 * 	\retval false reference count NOT decremented (no init, mutex, buffer_ref was not found or 
 * 				 not taken at all)
 * 
 * 	\note
 * 		This function contains assert() which gives more failure reasons during development/debugging.
 */
bool CNQMessagePool::putback_reference( CNQMessage *ref )
{
	if (m_bInitialized == false) return false;

	bool retval = false;

	if (!lock()) return retval;

	for (UINT32 i=0;i<m_iNumBuffers;i++)
	{
		if (&(m_pCMessages[i]) == ref)
		{
			assert( m_pTakenBuffer[i] >= 1);
			m_pTakenBuffer[i]--;

			// clear it if put back(!) important step
			if (m_pTakenBuffer[i] == 0) 
			{
				m_pCMessages[i].clear(); 
				assert(m_iSharedReferences > 0);
				m_iSharedReferences--;
			}
			
			retval = true;
		}
	}

	unlock();

	return retval;
}
