#include "PlatformSW.h"
#include <assert.h>
#include "NQDeclarations.h"
#include "NQPeerList.h"


//
// CNQPeerList constructor
//
//	initialization and memory allocation 
//
CNQPeerList::CNQPeerList() : CNQLockMechanism() 
{
	INT32 maxsize = NQ_PEERLIST_MAX_SIZE; // # entries 
	m_iQueueMaxSize = maxsize;

	m_pQueueValue = new NQ_PEER_ENTRY_INT[maxsize];
	m_iQueueNextElement = new INT32[maxsize];

	m_iLastError = ERR_PL_NONE;

	for (int i=0;i<maxsize;i++)
	{
		m_pQueueValue[i].enabled = false;
		m_iQueueNextElement[i] = -1;
	}

	m_iQueueFirstElement = -1;
	m_iQueueLastElement = -1;
	m_iQueueNumElements = 0;
}

//
// CNQPeerList destructor
//
//	mandatory deinitialization (free up memory)
//
CNQPeerList::~CNQPeerList() 
{
	delete [] m_pQueueValue;
	delete [] m_iQueueNextElement;
}



//
// convenience
//

bool CNQPeerList::add( UINT32 ip, UINT16 port )
{
	UINT32 drop_lasterr;
	
	return add(ip, port, drop_lasterr);
}

bool CNQPeerList::remove( UINT32 ip, UINT16 port )
{
	UINT32 drop_lasterr;

	return remove( ip, port, drop_lasterr );
}

bool CNQPeerList::removeAll()
{
	UINT32 drop_lasterr;

	return removeAll( drop_lasterr );
}

/*not used at the moment, untested, removed from release SW! Aug 28, 2007
bool CNQPeerList::find( UINT32 ip, UINT16 port)
{
	UINT32 drop_lasterr;

	return find( ip, port, drop_lasterr );
}*/

/*not used at the moment, untested, removed from release SW! Aug 28, 2007
bool CNQPeerList::getEntryByPos( UINT32 pos, UINT32 &ip, UINT32 &port)
{
	UINT32 drop_lasterr;

	return getEntryByPos(ip, port, drop_lasterr);
}*/

bool CNQPeerList::isAllowed(UINT32 ip, UINT16 port)
{
	UINT32 drop_lasterr;
	
	return isAllowed( ip, port, drop_lasterr);
}

bool CNQPeerList::getReceivePacketIndex( UINT32 ip, UINT16 port, UINT16 &packetindex)
{
	UINT32 drop_lasterr;
	
	return getReceivePacketIndex( ip, port, packetindex, drop_lasterr);
}

bool CNQPeerList::setReceivePacketIndex( UINT32 ip, UINT16 port, UINT16 packetindex)
{
	UINT32 drop_lasterr;
	
	return setReceivePacketIndex( ip, port, packetindex, drop_lasterr);
}

bool CNQPeerList::getSendPacketIndex( UINT32 ip, UINT16 port, UINT16 & aNextPacketIndex)
{
	UINT32 drop_lasterr;

	return getSendPacketIndex( ip, port, aNextPacketIndex, drop_lasterr );
}

bool CNQPeerList::incSendPacketIndex( UINT32 ip, UINT16 port)
{
	UINT32 drop_lasterr;
	
	return incSendPacketIndex( ip, port, drop_lasterr);
}

//
// convenience END
//

//
// CNQPeerList::add()
//
/** \brief adds a new peer to the list of peers.
 * 
 * 	\param ip       ip address of peer to add in host byte order
 *  \param port     network port of peer to add in host byte order
 *  \param lasterr	unique error code
 * 
 *  \retval true    successful, element appended OR element was already inside
 *  \retval false   failure, something happened (see lasterr)
 *
 *  \b Errorcodes:
 *   - #ERR_PL_LOCK - mutex locking failed
 *   - #ERR_PL_FULL	- queue is full
 * 
 * 	\note
 * 		This method returns true even in case the peer was already added to the list before. \p
 * 		This method contains assert() which gives more failure reasons during development/debugging.
 * 
 *  \sa nq_errorcodes
 */
bool CNQPeerList::add( UINT32 ip, UINT16 port, UINT32 &lasterr)
{
	if (!lock()) 
	{
		m_iLastError = lasterr = ERR_PL_LOCK;
		return false;
	}
	
	bool retval = false;
	
	INT32 it = m_iQueueFirstElement;
	
	if (m_iQueueNumElements != 0) 
	{
		while( it != -1)
		{
			assert( m_pQueueValue[it].enabled == true );

			if ((ip == m_pQueueValue[it].ip_host) && (port == m_pQueueValue[it].port_host))
			{
				retval = true;
				m_iLastError = lasterr = ERR_PL_NONE;
				break;
			}
		
			it = m_iQueueNextElement[it];
		}
	}
	
	if (retval == false) // remote peer is not yet in list
	{

		// if queue full: dont append anything else
		if (m_iQueueNumElements >= m_iQueueMaxSize) 
		{
			assert( m_iQueueNumElements == m_iQueueMaxSize );
			m_iLastError = lasterr = ERR_PL_FULL;
		}
		else
		{
			INT32 free_index = -1;

			for (INT32 cur=0;cur<m_iQueueMaxSize;cur++)
			{
				if (m_pQueueValue[cur].enabled == false) 
				{
					free_index = cur;
					break;
				}
			}

			assert( free_index != -1 );
		
			if (m_iQueueLastElement == -1)
			{
				// queue is empty
				m_iQueueFirstElement = free_index;
				m_iQueueLastElement = free_index;

				m_pQueueValue[free_index].enabled = true;
				m_pQueueValue[free_index].ip_host = ip;
				m_pQueueValue[free_index].port_host = port;
				m_pQueueValue[free_index].receive_last_packet_index_set = false;
				m_pQueueValue[free_index].send_packet_index = 1;

				m_iQueueNextElement[free_index] = -1;
				m_iQueueNumElements++;
			}
			else
			{
				m_iQueueNextElement[m_iQueueLastElement] = free_index;

				m_pQueueValue[free_index].enabled = true;
				m_pQueueValue[free_index].ip_host = ip;
				m_pQueueValue[free_index].port_host = port;
				m_pQueueValue[free_index].receive_last_packet_index_set = false;
				m_pQueueValue[free_index].send_packet_index = 1;

				m_iQueueNextElement[free_index] = -1;
				m_iQueueLastElement = free_index;
				m_iQueueNumElements++;
			}
			
			m_iLastError = lasterr = ERR_PL_NONE;
			retval = true;
		}
	}
			
	unlock();

	return retval;
}

//
// CNQPeerList::remove()
//
/** \brief Removes a peer from the list of peers.
 * 
 * 	\param ip       ip address of peer to remove in host byte order
 *  \param port 	network port of peer to remove in host byte order
 *  \param lasterr	unique error code
 * 
 * 	\retval true    successful, element removed
 *  \retval false	failure, something happened (see lasterr)
 * 					 - #ERR_PL_LOCK		- mutex locking failed
 * 					 - #ERR_PL_NOTFOUND	- peer element(ip, port) was NOT found
 * 
 * 	\note
 * 		This method contains assert() which gives more failure reasons during development/debugging.
 * 
 *  \sa nq_errorcodes
 */ 
bool CNQPeerList::remove( UINT32 ip, UINT16 port, UINT32 &lasterr )
{
	if (!lock()) 
	{
		m_iLastError = lasterr = ERR_PL_LOCK;
		return false;
	}
	
	INT32 found_index = m_iQueueFirstElement;
	INT32 last_index = -1;
	bool del = false;
	bool retval = false;

	while( found_index != -1)
	{
		assert( m_pQueueValue[found_index].enabled == true);

		if ((m_pQueueValue[found_index].ip_host == ip) && (m_pQueueValue[found_index].port_host == port)) 
		{
			if (last_index == -1)
			{
				INT32 removeidx = m_iQueueFirstElement;

				m_pQueueValue[removeidx].enabled = false;
				m_iQueueFirstElement = m_iQueueNextElement[removeidx];
				m_iQueueNumElements--;

				if (m_iQueueNumElements == 0)
				{
					assert( m_iQueueFirstElement == -1 );
					m_iQueueLastElement = -1;
				}

				del = true;
			}
			else
			{
				INT32 removeidx = found_index;

				m_pQueueValue[removeidx].enabled = false;
				m_iQueueNextElement[last_index] = m_iQueueNextElement[removeidx];
				m_iQueueNumElements--;
				m_iQueueNextElement[removeidx] = -1;

				if (m_iQueueNextElement[last_index] == -1)
				{
					m_iQueueLastElement = last_index;
				}

				del = true;

			}
		}
		
		if (del == true) break;

		last_index = found_index;
		found_index = m_iQueueNextElement[found_index];
	}

	if (del == true)
	{
		retval = true;
		m_iLastError = lasterr = ERR_PL_NONE;
	}
	else m_iLastError = lasterr = ERR_PL_NOTFOUND;


	unlock();

	return retval;
}

//
// CNQPeerList::removeAll()
//
/** \brief Clears the list of peers (makes it empty).
 * 
 * 	\param lasterr unique error code
 * 
 * 	\retval true	    successful, peer list is empty now
 * 	\retval false    failure, something happened (see lasterr)
 *          		 - #ERR_PL_LOCK - mutex locking failed
 *
 */ 
bool CNQPeerList::removeAll( UINT32 &lasterr )
{
	if (!lock()) 
	{
		m_iLastError = lasterr = ERR_PL_LOCK;
		return false;
	}


	for (int i=0;i<m_iQueueMaxSize;i++)
	{
		m_pQueueValue[i].enabled = false;
		m_iQueueNextElement[i] = -1;
	}

	m_iQueueFirstElement = -1;
	m_iQueueLastElement = -1;
	m_iQueueNumElements = 0;

	unlock();

	return true;
}


//
// CNQPeerList::count()
//
/** \brief Returns the number of elements (number of peers) in the peer list.
 * 
 *  \returns number of elements contained in the peer list
 */
UINT32 CNQPeerList::count()
{
	return( m_iQueueNumElements );

}

/* not used at the moment, untested, removed from release SW! Aug 28, 2007
bool CNQPeerList::getEntryByPos( UINT32 pos, UINT32 &ip, UINT16 &port, UINT32 &lasterr)
{
	if (!lock()) 
	{
		m_iLastError = lasterr = ERR_PL_LOCK;
		return false;
	}

	INT32 it = m_iQueueFirstElement;
	bool retval = false;
	UINT32 cnt=0;

	m_iLastError = lasterr = ERR_PL_NOTFOUND;

	if (m_iQueueNumElements == 0) return false;

	while( it != -1)
	{
		if (cnt == pos) 
		{
			assert( m_pQueueValue[it].enabled == true );

			ip = m_pQueueValue[it].ip_host;
			port = m_pQueueValue[it].port_host;
			retval = true;
			m_iLastError = lasterr = ERR_PL_NONE;
			break;
		}
		
		cnt++;
		it = m_iQueueNextElement[it];
    }
	
	unlock();

	return retval;
}
*/
/* not used at the moment, untested, removed from release SW! Aug 28, 2007
bool CNQPeerList::find( UINT32 ip, UINT16 port, UINT32 &lasterr)
{
	if (!lock()) 
	{
		m_iLastError = lasterr = ERR_PL_LOCK;
		return false;
	}

	INT32 it = m_iQueueFirstElement;
	bool retval = false;

	m_iLastError = lasterr = ERR_PL_NOTFOUND;

	if (m_iQueueNumElements != 0) 
	{
		while( it != -1)
		{
			assert( m_pQueueValue[it].enabled == true );

			if ((ip == m_pQueueValue[it].ip_host) && (port == m_pQueueValue[it].port_host))
			{
				retval = true;
				m_iLastError = lasterr = ERR_PL_NONE;
				break;
			}
		
			it = m_iQueueNextElement[it];
		}
	}
	
	unlock();

	return retval;
}
*/


//
// CNQPeerList::isAllowed()
//
/** \brief Checks whether a peer is in the list.
 * 
 * 	This method checks whether a peer (defined as (ip, port)) is inside the list (and thus, it is 
 *  allowed to receive packets from and send packets to the peer)
 * 
 *  \param ip       ip address of peer to check in host byte order
 * 	\param port		network port of peer to check in host byte order
 * 	\param lasterr 	unique error code
 * 
 * 	\retval true    successful, peer is in the list and thus allowed
 * 	\retval false   peer is not allowed; failure, something happened (see lasterr)
 * 					 - #ERR_PL_LOCK     - mutex locking failed
 * 					 - #ERR_PL_NOTFOUND	- peer not found
 * 
 * 	\note
 * 		This method returns false even in case the peer is valid, but mutex locking could not be performed.
 * 
 *  \sa nq_errorcodes
 */ 
bool CNQPeerList::isAllowed(UINT32 ip, UINT16 port, UINT32 &lasterr)
{
	if (!lock()) 
	{
		m_iLastError = lasterr = ERR_PL_LOCK;
		return false;
	}

	INT32 it = m_iQueueFirstElement;
	bool retval = false;

	m_iLastError = lasterr = ERR_PL_NOTFOUND;

	while( it != -1)
	{
		assert( m_pQueueValue[it].enabled == true );

		if ((ip == m_pQueueValue[it].ip_host) && (port == m_pQueueValue[it].port_host))
		{
			retval = true;
			m_iLastError = lasterr = ERR_PL_NONE;
			break;
		}
		
		it = m_iQueueNextElement[it];
	}
	
	unlock();

	return retval;
}

//
// CNQPeerList::getReceivePacketIndex()
//
/** Returns received packet index from given peer.
 * 
 * 	This method returns the last received packet index that came from the peer, IF there was already
 *  a packet received from that peer.
 * 
 * 	\param ip       ip address of peer in host byte order
 * 	\param port	    network port of peer in host byte order
 * 	\param packetindex  last packet index from peer (on true return code) 
 * 	\param lasterr		unique error code
 * 
 * 	\retval true    successful, packetindex is returned
 * 	\retval false   failure, something happened (see lasterr)
 * 					 - #ERR_PL_LOCK		   - mutex locking failed
 * 					 - #ERR_PL_NOTFOUND	   - peer not found in list
 * 					 - #ERR_PL_INDEXNOTSET - receive packet index is not yet set
 * 
 * 	\note
 * 		This method contains assert() which gives more failure reasons during development/debugging.
 * 
 *  \sa nq_errorcodes
 */ 
bool CNQPeerList::getReceivePacketIndex( UINT32 ip, UINT16 port, UINT16 &packetindex, UINT32 &lasterr)
{
	if (!lock()) 
	{
		m_iLastError = lasterr = ERR_PL_LOCK;
		return false;
	}

	INT32 it = m_iQueueFirstElement;
	bool retval = false;

	m_iLastError = lasterr = ERR_PL_NOTFOUND;

	while( it != -1)
	{
		assert( m_pQueueValue[it].enabled == true );

		if ((ip == m_pQueueValue[it].ip_host) && (port == m_pQueueValue[it].port_host))
		{
			if (m_pQueueValue[it].receive_last_packet_index_set == true)
			{
				packetindex = m_pQueueValue[it].receive_last_packet_index;
				retval = true;
				m_iLastError = lasterr = ERR_PL_NONE;
				break;
			}
			else
			{
				m_iLastError = lasterr = ERR_PL_INDEXNOTSET;
				break;
			}
		}
		
		it = m_iQueueNextElement[it];
	}
	
	unlock();

	return retval;
}

//
// CNQPeerList::setReceivePacketIndex()
//
/** Returns the last received packet index from given peer.
 * 
 * 	This method returns the last received packet index that came from the peer, IF there was already
 *   a packet received from that peer.
 * 
 * 	\param ip       ip address of peer in host byte order
 * 	\param port		network port of peer in host byte order
 * 	\param packetindex  last packet index from peer (to save inside peerlist) 
 * 	\param lasterr  unique error code
 * 
 * 	\retval true    successful, packetindex was updated inside peerlist
 * 	\retval false   failure, something happened (see lasterr)
 * 					 - #ERR_PL_LOCK		- mutex locking failed
 * 					 - #ERR_PL_NOTFOUND	- peer not found in list
 * 
 *  \note
 * 		This method contains assert() which gives more failure reasons during development/debugging.
 * 
 *  \sa nq_errorcodes
 */ 
bool CNQPeerList::setReceivePacketIndex( UINT32 ip, UINT16 port, UINT16 packetindex, UINT32 &lasterr)
{
	if (!lock()) 
	{
		m_iLastError = lasterr = ERR_PL_LOCK;
		return false;
	}

	INT32 it = m_iQueueFirstElement;
	bool retval = false;

	m_iLastError = lasterr = ERR_PL_NOTFOUND;

	while( it != -1)
	{
		assert( m_pQueueValue[it].enabled == true );

		if ((ip == m_pQueueValue[it].ip_host) && (port == m_pQueueValue[it].port_host))
		{
			m_pQueueValue[it].receive_last_packet_index = packetindex;
			m_pQueueValue[it].receive_last_packet_index_set = true;
			retval = true;
			m_iLastError = lasterr = ERR_PL_NONE;
			break;
		}
		
		it = m_iQueueNextElement[it];
	}
	
	unlock();

	return retval;
}


//
// CNQPeerList::getSendPacketIndex()
//
/** \brief Returns the index of the last packet that was send to the peer.
 * 
 * 	\param ip       ip address of peer in host byte order
 *  \param port		network port of peer in host byte order
 * 	\param aNextPacketIndex last send packet index to peer (on true return code) 
 * 	\param lasterr	unique error code
 * 
 * 	\retval	true    successful, last send packet index is returned
 * 	\retval false   failure, something happened (see lasterr)
 * 					 - #ERR_PL_LOCK		- mutex locking failed
 * 					 - #ERR_PL_NOTFOUND - peer not found in list
 * 
 *  \note
 * 		This method contains assert() which gives more failure reasons during development/debugging.
 * 
 *  \sa nq_errorcodes
 */ 
bool CNQPeerList::getSendPacketIndex( UINT32 ip, UINT16 port, UINT16 & aNextPacketIndex, UINT32 &lasterr )
{
	if (!lock()) 
	{
		m_iLastError = lasterr = ERR_PL_LOCK;
		return false;
	}

	INT32 it = m_iQueueFirstElement;
	bool retval = false;

	m_iLastError = lasterr = ERR_PL_NOTFOUND;

	while( it != -1)
	{
		assert( m_pQueueValue[it].enabled == true );

		if ((ip == m_pQueueValue[it].ip_host) && (port == m_pQueueValue[it].port_host))
		{
			aNextPacketIndex = m_pQueueValue[it].send_packet_index;
			retval = true;
			m_iLastError = lasterr = ERR_PL_NONE;
			break;
		}
		
		it = m_iQueueNextElement[it];
	}
	
	unlock();

	return retval;
}

//
// CNQPeerList::incSendPacketIndex()
//
/** \brief Increments the packet index for given peer
 * 
 * 	This method increments the send packet index of the given peer (because a packet was successfully send out 
 *   by the network queue) by one.
 * 
 * 	\param ip       ip address of peer in host byte order
 * 	\param port		network port of peer in host byte order
 * 	\param lasterr  unique error code
 * 
 * 	\retval	true    successful, send packet index was incremented (by one)
 * 	\retval false   failure, something happened (see lasterr)
 * 					 - #ERR_PL_LOCK - mutex locking failed
 * 					 - #ERR_PL_NOTFOUND - peer not found in list
 * 
 * 	\note
 * 		This method contains assert() which gives more failure reasons during development/debugging.
 * 
 *  \sa nq_errorcodes
 */ 
bool CNQPeerList::incSendPacketIndex( UINT32 ip, UINT16 port, UINT32 &lasterr)
{
	if (!lock()) 
	{
		m_iLastError = lasterr = ERR_PL_LOCK;
		return false;
	}

	INT32 it = m_iQueueFirstElement;
	bool retval = false;

	m_iLastError = lasterr = ERR_PL_NOTFOUND;

	while( it != -1)
	{
		assert( m_pQueueValue[it].enabled == true );

		if ((ip == m_pQueueValue[it].ip_host) && (port == m_pQueueValue[it].port_host))
		{
			m_pQueueValue[it].send_packet_index++;

			retval = true;
			m_iLastError = lasterr = ERR_PL_NONE;
			break;
		}
		
		it = m_iQueueNextElement[it];
	}
	
	unlock();

	return retval;
}

/** \brief Returns last error code happened
 * 
 *  For detailed list of possible error codes, see Declarations.h
 * 
 *  \sa nq_errorcodes
 */ 
UINT32 CNQPeerList::getLastError()
{
	// be very careful: concurrent threads
	return( m_iLastError );
}
